class Embed:
    """
    A Discord embed.

    Attributes:
        title (str): Title of the embed
        embed_type (str): An embed type (either 'rich', 'image', 'video', 'gifv', 'article', or 'link')
        description (str): Embed description
        url (str): URL provided in embed
        timestamp (str): An ISO8601-compliant timestamp string
        color (int): The color for the embed, better represented as a hexadecimal value.
        colour (int): Embed colour
        fields (list): Embed fields
    """

    def __init__(self, data: dict):
        self.title: str = data.get("title", "")
        self.embed_type: str = data.get("type", "rich")

        self.description: str = data.get("description", "")
        self.url: str = data.get("url", "")
        self.timestamp: str = data.get("timestamp", "")

        self.color: int = data.get("color", "")

        self.colour: int = self.color
        self.fields: list = []

        # Privates
        self.__thumbnail = {}
        self.__image = {}
        self.__video = {}
        self.__footer = {}
        self.__author = {}
        self.__provider = {}

    @classmethod
    def create(cls, *, title: str = None, description: str = None,
               url: str = None, color: int = None
               ):
        """
        Create a new Embed object

        Args:
            title (str): (Optional) Title of the embed
            description (str): (Optional) Embed description
            url (str): (Optional) URL provided in Embed
            color (int): (Optional) Color of embed
        """

        return Embed({
            "title": title,
            "description": description,
            "url": url,
            "color": color
        })

    def add_field(self, name: str, value: str, inline: bool = True):
        """
        dd a field object to the embed fields
        
        Args:
            name (str): The name of the embed field
            value (str): The contents/value of the field
            inline (bool): Indicates whether the field is placed either inline or in a new row. \
                True by default
        """

        self.fields.append({
            "name": name,
            "value": value,
            "inline": inline
        })

    def __make_image(self, url, proxy_url, height, width):
        return {
            "url": url,
            "proxy_url": proxy_url,
            "height": height,
            "width": width
        }

    @property
    def thumbnail(self):
        """Embed thumbnail"""
        return self.__thumbnail

    def set_thumbnail(self, url: str, *, proxy_url: str = None, height: int = None, width: int = None):
        """
        Add a thumbnail to an embed object.

        Args:
            url (str): The URL for the thumbnail
            proxy_url (str): An alternative URL for the thumbnail
            height (int): The height of the thumbnail
            width (int): The width of the thumbnail
        """

        self.__thumbnail = self.__make_image(url, proxy_url, height, width)

    @property
    def image(self):
        """Embed image"""
        return self.__image

    def set_image(self, url, *, proxy_url: str = None, height: int = None, width: int = None):
        """
        Set embed image.

        Args:
            url (str): The URL for the image
            proxy_url (str): An alternative URL for the image
            height (int): The height of the image
            width (int): The width of the image
        """

        self.__image = self.__make_image(url, proxy_url, height, width)

    @property
    def video(self):
        """Embed video"""
        return self.__video

    def set_video(self, url, *, proxy_url: str = None, height: int = None, width: int = None):
        """
        Set embed video.

        Args:
            url (str): The URL for the video
            proxy_url (str): An alternative URL for the video
            height (int): The height of the video
            width (int): The width of the video
        """
        self.__video = self.__make_image(url, proxy_url, height, width)

    ##: Footer
    @property
    def footer(self):
        """Embed footer"""
        return self.__footer

    def set_footer(self, *, text: str, icon_url: str, proxy_url: str = None):
        """
        Set embed footer.

        Args:
            text (str): The text for the footer
            icon_url (str): An URL for the footer
            proxy_url (str): An alternative URL for the footer
        """

        self.__footer = {
            "text": text,
            "icon_url": icon_url,
            "proxy_icon_url": proxy_url
        }

    ##: Author
    @property
    def author(self):
        """Author of embed"""
        return self.__author

    def set_author(self, *, name: str, url: str, icon_url: str, proxy_url: str = None):
        """
        Set the author of an embed.

        Args:
            name (str): The name of the author
            url (str): An URL related to the author
            icon_url (str): An icon URL for the author
            proxy_url (str): A proxy URL for the author
        """

        self.__author = {
            "name": name,
            "url": url,
            "icon_url": icon_url,
            "proxy_icon_url": proxy_url
        }

    @property
    def provider(self):
        """Embed Provider"""
        return self.__provider

    def set_provider(self, *, name: str = None, url: str = None):
        """
        Set the provider of the embed.

        Args:
            name (str): The name of the provider
            url (str): The URL of the provider
        """
        self.__provider = {
            "name": name,
            "url": url
        }

    def to_dict(self):
        """Convert object into a dictionary"""
        return {
            "title": self.title,
            "type": self.embed_type,
            "description": self.description,
            "url": self.url,
            "timestamp": self.timestamp,
            "color": self.color,
            "fields": self.fields,
            "thumbnail": self.thumbnail,
            "image": self.image,
            "video": self.video,
            "footer": self.footer,
            "author": self.author,
            "provider": self.provider
        }

    def __repr__(self):
        return f"<Embed title='{self.title}' description='{self.description}'>"



from __future__ import annotations
from datetime import datetime

from pycordia import api, utils
from typing import List, Optional, Union
import typing
import aiohttp
import enum

import pycordia

from . import embed
from .user import User
from .guild import Member, Role, Emoji
from .channel import ChannelMention
from pycordia import models


class MessageActivity:
    """
    Message Activity class

    Attributes:
        activity_type (`pycordia.models.message.MessageActivity.ActivityTypes`): Type of activity
        party_id (str): ID of the party
    """

    class ActivityTypes(enum.Enum):
        join = 1
        spectate = 2
        listen = 3
        join_request = 5

    def __init__(self, data: dict):
        Activity = MessageActivity.ActivityTypes
        self.activity_type = (
            None if (data.get("type") is None) else Activity(data.get("type"))
        )

        self.party_id: Optional[str] = data.get("party_id")

    def __repr__(self):
        return f"<MessageActivity id={self.party_id} activity={self.activity_type}>"


class Application:
    """Discord application model."""
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.icon_hash: Optional[str] = data.get("icon")
        self.description: str = data["description"]

        self.rpc_origins: list = data.get("rpc_origins", [])
        self.bot_public: bool = data["bot_public"]
        self.bot_require_code_grant: bool = data["bot_require_code_grant"]

        self.terms_of_service_url: Optional[str] = data.get("terms_of_service_url")
        self.privacy_policy_url: Optional[str] = data.get("privacy_policy_url")
        self.owner: Optional[User] = utils.make_optional(User, data.get("owner", {}))

        self.cover_image_hash: Union[str, None] = data.get("cover_image")
        self.flags: Union[int, None] = data.get("flags")

    def __repr__(self):
        return f"<Application id={self.id} name='{self.name}'>"

    @property
    def icon_url(self) -> Optional[str]:
        """The URL for this application's icon"""
        if self.icon_hash:
            return f"{pycordia.cdn_url}/app-icons/{self.id}/{self.icon_hash}.png"

    @property
    def cover_image_url(self) -> Optional[str]:
        """The URL for this application's cover image"""
        if self.cover_image_hash:
            return f"{pycordia.cdn_url}/app-icons/{self.id}/{self.cover_image_hash}.png"


class Reaction:
    """Represents reactions in a message."""
    def __init__(self, data: dict):
        self.count: Optional[int] = data.get("count")
        self.me: Optional[bool] = data.get("me")
        self.emoji: Optional[Emoji] = utils.make_optional(Emoji, data.get("emoji", {}))

    def __repr__(self):
        return f"<Reaction emoji={str(self.emoji)} count={self.count}"



class StickerItem:
    """Represents a Discord sticker item"""
    def __init__(self, data: dict):
        self.sticker_id = data.get("id")
        self.name = data.get("name")
        self.format_type = data.get("type")

    def __repr__(self):
        return f"<StickerItem id={self.sticker_id} name='{self.name}'>"


class Attachment:
    """
    An attachment for a message

    Attributes:
        attachment_id (int): The ID of the attachment
        filename (str): The attachment's filename
        content_type (str): The content type
        size (int): File size
        url (str): URL to the file
        proxy_url (str): Alternate URL to the file
        height (int): Height of attachment (only applies to images)
        width (int): Width of the attachment (only applies to images)
    """
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.filename: str = data["filename"]
        self.content_type: Optional[str] = data.get("content_type")
        self.size: Optional[int] = data.get("size")
        self.url: Optional[str] = data.get("url")
        self.proxy_url: Optional[str] = data.get("proxy_url")
        self.height: Optional[int] = data.get("height")
        self.width: Optional[int] = data.get("width")

    def to_dict(self):
        """
        Return raw dictionary of object
        Returns: dict
        """
        return {
            "id": self.id,
            "filename": self.filename,
            "content_type": self.content_type,
            "size": self.size,
            "url": self.url,
            "proxy_url": self.proxy_url,
            "height": self.height,
            "width": self.width,
        }


class Interaction:
    """
    Represents a Discord interaction

    Attributes:
        interaction_id (str): ID of the discord interaction
        interaction_type (str): Type of the interaction (button, slash command, etc)
        name (str): Name of the interaction
        user (`pycordia.models.user.User`): Interaction's user
    """
    def __init__(self, data: dict):
        self.interaction_id = data.get("id")
        self.interaction_type = data.get("type")
        self.name = data.get("name")
        self.user = User(data.get("user", {}))


class MessageReference:
    def __init__(self, ref_data: dict, msg_data: dict):
        self.__msg_data = msg_data

        if ref_data:
            self.message_id: Optional[str] = ref_data.get("message_id")
            self.channel_id: Optional[str] = ref_data.get("channel_id")
            self.guild_id: Optional[str] = ref_data.get("guild_id")
            self.fail_if_not_exists: Optional[bool] = ref_data.get(
                "fail_if_not_exists"
            )
        else:
            self.message_id = (
                self.channel_id
            ) = self.guild_id = self.fail_if_not_exists = None

    @property
    def message(self):
        """The referenced message"""
        return utils.make_optional(Message, self.__msg_data)

    def to_dict(self):
        return {
            "message_id": self.message_id,
            "channel_id": self.channel_id,
            "guild_id": self.guild_id,
            "fail_if_not_exists": self.fail_if_not_exists,
        }
    
    def __eq__(self, other) -> bool:
        return self.message_id == other.message_id


class Message:
    """Represents a message sent in Discord
    
    Attributes:
        id (str): ID of the message
        channel_id (str): ID of the channel the message was sent in
        guild_id (str): ID of the guild the message was sent in
        author (User): The author of this message
        member (Member): The author of this message, including guild information
        content (str): Content of the message
        tiemstamp (datetime): The time the message was sent
        edited_timestamp (datetime): The time the message was edited
        tts (bool): Whether the message supports Text to Speech
        mention_everyone (bool): Whether this message mentions everyone
        mentions (List[User]): A list of users this message mentions
        mention_roles (List[Role]): A list of roles this message mentions
        mention_channels (List[ChannelMention]): A list of channels this message mentions
        attachments (List[Attachment]): The attachments for this message
        embeds (List[embed.Embed]): The embeds for this message
        reactions (List[Reaction]): Reactions for this message
        nonce (Union[str, int, None]): Used for message verification
        pinned (bool): Whether this message is pinned
        webhook_id (str): The ID Of the webhook that sent this message
        type (int): Message type;
        activity (str): The Rich-Presence chat embed if any
        application (str): The application for the Rich-Presence chat embed
        application_id (str): The ID of the application for the embed
        message_reference (MessageReference): The message referred to
        flags (int): Flags for this message
        interaction (dict): An interaction for this message -- not implemented
        thread (dict): The thread that was started from this message -- not implemented
        components (dict): The list of components for this message -- not implemented
        sticker_items (List[StickerItem]): Stickers sent in this message
    """
    def __init__(self, data: dict):
        self.__data = data

        self.id: str = data["id"]
        self.channel_id: str = data["channel_id"]
        self.guild_id: Optional[str] = data.get("guild_id")
        self.author: User = User(data["author"])

        self.member: Optional[Member] = utils.make_optional(Member, data.get("member", {}))
        self.content: str = data["content"]
        self.timestamp: datetime = datetime.fromisoformat(data["timestamp"])
        self.edited_timestamp: Optional[datetime] = utils.make_optional(datetime.fromisoformat, data.get("edited_timestamp"))
        self.tts: bool = data["tts"]

        self.mention_everyone: bool = data["mention_everyone"]

        self.mentions: List[User] = list(map(User, data.get("mentions", [])))
        self.mention_roles: List[Role] = list(map(Role, data.get("mention_roles", [])))
        self.mention_channels: List[ChannelMention] = list(
            map(ChannelMention, data.get("mention_channels", []))
        )

        self.attachments: List[Attachment] = list(map(Attachment, data.get("attachments", [])))
        self.embeds: List[embed.Embed] = list(map(embed.Embed, data.get("embeds", [])))
        self.reactions: List[Reaction] = list(map(Reaction, data.get("reactions", [])))
        self.nonce: Union[str, int, None] = data.get("nonce")
        self.pinned: bool = data["pinned"]
        self.webhook_id: Optional[str] = data.get("webhook_id")
        self.type: int = data["type"]
        self.activity: Optional[MessageActivity] = utils.make_optional(MessageActivity, data.get("activity", {}))

        self.application: Optional[Application] = utils.make_optional(Application, data.get("application", {}))
        self.application_id: Optional[str] = data.get("application_id")
        self.message_reference: MessageReference = MessageReference(
            data.get("message_reference", {}), 
            data.get("referenced_message", {})
        )

        self.flags: Optional[int] = data.get("flags")

        self.interaction: dict = data.get("interaction", {})
        self.thread: dict = data.get("thread", {})
        self.components: list = data.get("components", [])

        self.sticker_items: List[StickerItem] = list(map(StickerItem, data.get("sticker_items", [])))

    @classmethod
    async def send(cls, channel_id: str, *,
        content: str = "", 
        embeds: List[embed.Embed] = None,
        allowed_mentions: dict = {},
        allow_tts: bool = False
    ) -> Message:
        """Send a Discord message to a channel

        Args:
            channel_id (str): The ID of the channel where the message will be sent
            content (str, optional): The content of the message.
            embeds (List[embed.Embed], optional): A list of `pycordia.models.Embed` objects.
            allowed_mentions (dict, optional): \
                A dictionary of allowed mentions. Dictionary keys are as follows, \
                    dictionary values are booleans: \n
                roles: Allow role mentions \n
                users: Allow user mentions \n
                everyone: Allow @everyone and @here mentions \n
            allow_tts (bool, optional): Indicate if the message allows text-to-speech. 

        Returns: `pycordia.models.Message`
        """
        # NOTE: Future use
        # if reply_to:
        #     reply = reply_to.to_dict()
        # else:
        #     reply = None
        rs = await api.request("POST", f"channels/{channel_id}/messages", json_data={
            "content": content or "",
            "tts": allow_tts,
            "allowed_mentions": allowed_mentions,
            "embeds": [emb.to_dict() for emb in (embeds or [])],
        })
        return Message(rs)

    @classmethod
    async def from_id(cls, channel_id: str, message_id: str) -> Message:
        """Get a message given a channel ID and a message ID
        
        Returns: `pycordia.models.Message`
        """
        return Message(
            await api.request("GET", f"channels/{channel_id}/messages/{message_id}")
        )

    async def delete(self):
        """Removes this message. Note that this action cannot be undone."""
        await api.request("DELETE", f"channels/{self.channel_id}/messages/{self.id}")

    # TODO: Implement attachments and files
    async def edit(self, *, content: str = None, 
        embeds: typing.List[models.Embed] = None,
        allowed_mentions: dict = None
    ):
        """Edit this message

        Args:
            content (str, optional): The new content of this message.
            embeds (typing.List[models.Embed], optional): A new list of `embed.Embed` objects 
            allowed_mentions (dict, optional): A new dictionary of allowed mentions. \
                Structure is the same as specified in `pycordia.models.Message.send`
        """
        await api.request("PATCH", 
            f"channels/{self.channel_id}/messages/{self.id}", 
            json_data={
                "content": content or self.content,
                "embeds": [embed.to_dict() for embed in (embeds or self.embeds)],
                "allowed_mentions": allowed_mentions
            }
        )      

    async def pin(self):
        """Pins a message to the channel it was sent in"""
        await api.request("PUT", f"channels/{self.channel_id}/pins/{self.id}")
    
    async def unpin(self):
        """Unpins a message from the channel it was sent in"""
        await api.request("DELETE", f"channels/{self.channel_id}/pins/{self.id}")



from datetime import datetime
import enum
from pycordia import api, utils
from typing import Dict, List, Union, Any, Optional

import aiohttp
import pycordia


class Channel:
    """
    Channel model to mirror a Discord Channel

    Operations:
        - str(x): Returns the channel name following a `#`
        - x == y: Checks if two channels are the same
    """
    class ChannelType(enum.Enum):
        text = 0
        dm = 1
        voice = 2
        group_dm = 3
        category = 4
        news = 5
        store = 6
        news_thread = 10
        public_thread = 11
        private_thread = 12
        stage_voice = 13

    def __init__(self, data: dict):
        self.__client = pycordia.models.active_client

        self.id: str = data["id"]
        self.type: Channel.ChannelType = Channel.ChannelType(data["type"])
        self.guild_id: Optional[str] = data.get("guild_id")
        self.position: Optional[int] = data.get("position")

        # TODO: Define overwrite model
        self.permission_overwrites: List[Dict] = data.get("permission_overwrites", [])

        self.name: Optional[str] = data.get("name")
        self.topic: Optional[str] = data.get("topic")
        self.nsfw: Optional[bool] = data.get("nsfw")
        self.last_message_id: Optional[str] = data.get("last_message_id")
       
        # -- Voice channel
        self.bitrate: Optional[int] = data.get("bitrate")
        self.user_limit: Optional[int] = data.get("user_limit")
        self.rate_limit_per_user: Optional[int] = data.get("rate_limit_per_user")

        # -- DMs
        self.recipients: list = data.get("recipients", [])
        
        self.icon_hash: Optional[str] = data.get("icon")

        self.owner_id: Optional[str] = data.get("owner_id")
        self.application_id: Optional[str] = data.get("application_id")
        self.parent_id: Optional[str] = data.get("parent_id")

        self.last_pin_timestamp: Optional[datetime] = utils.make_optional(datetime.fromisoformat, data.get("last_pin_timestamp"))
        self.rtc_region: Optional[str] = data.get("rtc_region")
        self.video_quality_mode: Optional[int] = data.get("video_quality_mode")

        self.message_count: int = data.get("message_count", -1)  # -1 if the count was not provided
        self.member_count: int = data.get("member_count", -1)


        # Thread specific stuff
        # TODO: Define models for these types
        self.thread_metadata: Dict[str, Any] = data.get("thread_metadata", {})
        self.thread_member: Dict[str, Any] = data.get("member", {})
        self.thread_default_auto_archive_duration: int = data.get("default_auto_archive_duration", -1)

        self.permissions: Optional[str] = data.get("permissions")

    def __repr__(self):
        return f"<Channel {self.type} id={self.id} name='{self.name}' topic='{self.topic}'>"

    def __str__(self):
        return f"#{self.name}"
    
    def __eq__(self, channel) -> bool:
        return self.guild_id == channel.guild_id and self.id == channel.id

    @property
    def mention(self) -> str:
        """
        A channel mention

        Returns: str
        """
        return f"<#{self.id}>"

    @classmethod
    async def from_id(cls, channel_id: str):
        """
        Fetch a channel object from its ID.

        Args:
            channel_id (str): Channel ID

        Returns: `pycordia.models.channel.Channel`
        """
        return Channel(await api.request("GET", f"channels/{channel_id}"))

    # TODO: Make this method support cache
    async def get_message(self, message_id: str) -> 'pycordia.models.Message':
        """ 
        Get message by ID

        Args:
            message_id (str): The message ID as a string

        Raises: `Exception` if no initialized client is found
        Returns: A `pycordia.models.Message` object
        """
        return pycordia.models.Message(
            await api.request("GET", f"channels/{self.id}/messages/{message_id}")
        )


    async def get_messages(self, limit: int = 50, *, 
        before: str = None, around: str = None, after: str = None
    ) -> List['pycordia.models.Message']:
        """ 
        Get the most recent messages, given a limit (by default 50)

        Args:
            limit (int): The amount of recent messages to get, \
                can be between 1 and 100 (by default: 50)
            before (str): Get messages before the message ID, as a string
            around (str): Get messages around the message ID, as a string
            after (str): Get messages after the message ID, as a string

        Raises: `Exception` if no initialized client is found
        Returns: A list of `pycordia.models.Message` objects
        """
        param_string = f"?limit={limit}"

        if before: param_string += f"&before={before}"
        if around: param_string += f"&around={around}"
        if after:  param_string += f"&after={after}"
        
        rs = await api.request("GET", f"channels/{self.id}/messages{param_string}")
        return list(map(pycordia.models.Message, rs))

    async def get_pinned_messages(self) -> List['pycordia.models.Message']:
        """ 
        Get the most recent messages, given a limit (by default 50)

        Raises: `Exception` if no initialized client is found
        Returns: A list of `pycordia.models.Message` objects
        """
        rs = await api.request("GET", f"channels/{self.id}/pins/")
        return list(map(pycordia.models.Message, rs))


class ChannelMention:
    """
    Represents a channel mention i.e. <#channel_id>

    Attributes:
        channel_id (str): ID of the channel
        guild_id (str): ID of the channel's guild
        channel_type (str): Type of the channel
        channel_name (str): Name of channel
    """
    def __init__(self, data: dict):
        self.channel_id: str = data["id"]
        self.guild_id: str = data["guild_id"]
        self.channel_type: str = data["type"]
        self.channel_name: str = data["name"]
    
    async def get_channel(self) -> Channel:
        """
        Fetch the corresponding channel object.

        Returns: `pycordia.models.channel.Channel`
        """
        return await Channel.from_id(self.channel_id)
        


from pycordia import api
import aiohttp
import asyncio
import enum

import typing

from . import events, models, websocket
import pycordia


class Intents(enum.Enum):
    guilds = 1 << 0
    guild_members = 1 << 1
    guild_bans = 1 << 2
    guild_emojis_and_stickers = 1 << 3
    guild_integrations = 1 << 4
    guild_webhooks = 1 << 5
    guild_invites = 1 << 6
    guild_voice_states = 1 << 7
    guild_presences = 1 << 8
    guild_messages = 1 << 9
    guild_message_reactions = 1 << 10
    guild_message_typing = 1 << 11
    direct_messages = 1 << 12
    direct_message_reactions = 1 << 13
    direct_message_typing = 1 << 14

    @classmethod
    def all(cls):
        """All registered intents, including privileged ones"""
        return cls.merge_intents(cls, True)

    @classmethod
    def standard(cls):
        """All non-privileged intents"""
        return cls.merge_intents(cls, False)

    @classmethod
    def merge_intents(cls, intent_list, privileged: bool = False):
        """Convert a list of intents into a number

        Parameters
        ---
            intent_list: A list of intents
            privileged (bool): Whether or not to include privileged \
                intents in the result
        """
        result = 0
        for value in intent_list:
            if not privileged and (value in (
                pycordia.Intents.guild_members, pycordia.Intents.guild_presences)): continue
            result |= value.value
        return result


class Client:
    """
    A WebSockets client for the Discord Gateway API

    Attributes:
        cache_size (int): Maz size of user and message caches
        message_cache (Dict[str, Message]): Client's message cache - a dictionary of string - `pycordia.models.message.Message` mappings
        user_cache (Dict[str, Message]): Client's user cache - a dictionary of string - `pycordia.models.user.User` mappings
    """

    def event(self, fun):
        self.register_events(fun.__name__, fun)

        def wrapper():
            fun()

        return wrapper

    def on(self, event_name):
        def factory(fun):
            self.register_events(event_name, fun)
            def wrapper():
                fun()
            return wrapper
        return factory

    def register_events(self, event_name: str, *callables: typing.Callable):
        """Register a set of callableeves for an event
        
        Args:
            event_name (str): The name of an event (example: `message_create` or `guild_create`)
            callables: A set of callable objects to call when `event_name` is triggered
        """
        name = f"{'on_'*(not event_name.startswith('on_'))}{event_name}"

        event = self.events.get(name)
        
        if event and isinstance(event, list):
            event.extend(callables)
        else:
            self.events[name] = [*callables]


    def __init__(self, *, intents: int, cache_size: int = 1000):
        """
        Args:
            intents (int): The intents for the bot to authenticate with. \
                Preferably obtained from `pycordia.Intents`
            cache_size (int, optional): The amount of entries \
                to store in cache at a time. Defaults to 1000.
        """
        self.events = {}
        self.__bot_token: str = ""
        self.intents = intents

        self.cache_size = int(cache_size)
        self.user_cache: typing.Dict[str, models.User] = {}
        self.message_cache: typing.Dict[str, models.Message] = {}

    async def __create_ws(self, bot_token):        
        if not bot_token:
            bot_token = ""

        self.ws = websocket.DiscordWebSocket(self, bot_token, self.intents)
        await self.ws.listen()

    async def call_event_handler(self, event_name: str, event_data):
        func_name = f"on_{event_name.lower()}"
        
        def call_handlers(*args, **kwargs):
            for event in self.events[func_name]:
                asyncio.gather(event(*args, **kwargs))

        # --- Cached methods ---
        if event_name.lower() == "message_create":
            message = models.Message(event_data)

            if len(self.message_cache.keys()) >= self.cache_size:
                first_message = list(self.message_cache.keys())[0]
                del self.message_cache[first_message]

            self.message_cache[message.id] = message

            if func_name in self.events:
                call_handlers(message)
                return
        elif event_name.lower() == "message_update":
            after = models.Message(event_data)
            
            # print(self.message_cache, after.id)
            
            before = self.message_cache.get(after.id, None)

            # Update the message cache
            if len(self.message_cache.keys()) >= self.cache_size:
                first_message = list(self.message_cache.keys())[0]
                del self.message_cache[first_message]
            
            self.message_cache[after.id] = after

            if func_name in self.events:
                call_handlers(before, after)               
                return

        # --- Uncached methods ---
        if func_name in self.events:
            func = self.events[func_name]

            if event_name.lower() == "ready":
                call_handlers(events.ReadyEvent(event_data))

            # ---- User Related Events ----
            elif event_name.lower() == "typing_start":
                call_handlers(events.TypingStartEvent(event_data))

            # ---- Message Related Events ----
            elif event_name.lower() == "message_delete":
                msg = self.message_cache.get(event_data["id"])

                call_handlers(
                    events.MessageDeleteEvent(event_data, False, [msg] if msg else [])
                )
            elif event_name.lower() == "message_delete_bulk":
                ids = event_data.get("ids", [])
                call_handlers(
                    events.MessageDeleteEvent(event_data, True, 
                        list(filter(lambda key: key in ids, self.message_cache))
                ))
            
            # ---- Channel Related Events ----
            elif event_name.lower() in ("channel_create", "channel_update", "channel_delete"):
                call_handlers(models.Channel(event_data))

            # ---- Unimplemented ----
            else:
                call_handlers(event_data)

    def run(self, bot_token: str):
        """Log into Discord, and start the event loop.

        ---        
        Parameters:
            bot_token (str): Discord token
        """
        self.__bot_token = bot_token
        pycordia.models.active_client = self

        loop = asyncio.get_event_loop()
        task = loop.create_task(self.__create_ws(bot_token))

        try:
            loop.run_until_complete(task)
        except KeyboardInterrupt:
            task.cancel()

            if self.ws.sock:
                loop.run_until_complete(self.ws.sock.close())

    async def get_channel(self, channel_id: str) -> models.Channel:
        return await models.Channel.from_id(channel_id)

    async def get_user(self, user_id: str, *, use_cache: bool = True) -> models.User:
        return await models.User.from_id(user_id, use_cache)

    async def get_guild(self, guild_id: str, *, with_counts: bool = False) -> models.Guild:
        return await models.Guild.from_id(guild_id, with_counts=with_counts)

    @property
    async def guilds(self) -> typing.List[models.PartialGuild]:
        """List of guilds of which the bot is a member"""
        rs = await api.request("GET", "users/@me/guilds")
        return list(map(models.PartialGuild, rs))

    @property
    async def user(self) -> models.User:
        """Bot's `pycordia.models.user.User` object"""
        return models.User(await api.request("GET", "users/@me"))

    # TODO: Create DM
    # TODO: Get User Connections



"""
Pycordia - The Discord bot framework
------------------------------------

Developed by Angel Carias and its contributors. 2021.

A wrapper around the Discord HTTP API and WebSockets
"""

from .client import Intents, Client
from . import websocket, errors, events, models, api

__version__ = "0.2.1"

cdn_url = "https://cdn.discordapp.com"
ws_url = "wss://gateway.discord.gg"
api_url = "https://discord.com/api/v9"



from typing import Optional
from pycordia import api, utils
from datetime import datetime
import pycordia


class User:
    """
    A Discord User

    Attributes:
        id (str): ID of the user
        username (str): A username
        discriminator (str): A user's discriminator; a `#` followed by 4 numbers
        avatar_hash (str): User's avatar hash
        bot (bool): Whether the user is a bot or not
        system (bool): Whether the user is an official Discord System User
        mfa_enabled (bool): Whether the user has 2FA enabled
        banner_hash (str): User's banner hash
        accent_color (int): User's banner color, better represented as hexadecimal
        locale (str): User's chosen language
        verified (bool): Whether the user's email has been verified. Can only be accessed from the email scope, in OAuth2 applications
        email (str): User's email. Can only be accessed from the email scope, in OAuth2 applications
        flags (int): Flags on user's account
        premium_type (int): Nitro subscription type: 0 - None, 1 - Nitro Classic, 2 - Nitro
        public_flags (int): Public flags on a user's account

    Operations:
        - str(x): Returns the username + discriminator
        - repr(x): Returns a representation of the object
        - x == y: Checks if two users are the same (check by ID)
    """
    def __init__(self, data):
        self.id: str = data["id"]
        self.username: str = data["username"]
        self.discriminator: str = data["discriminator"]
        self.avatar_hash: str = data["avatar"]
        self.bot: Optional[bool] = data.get("bot")
        self.system: Optional[bool] = data.get("system")
        self.mfa_enabled: Optional[bool] = data.get("mfa_enabled")
        self.banner_hash: Optional[str] = data.get("banner")
        self.accent_color: Optional[int] = data.get("accent_color")
        self.locale: Optional[str] = data.get("locale")
        self.verified: Optional[bool] = data.get("verified")
        self.email: Optional[str] = data.get("email")
        self.flags: Optional[int] = data.get("flags")
        self.premium_type: Optional[int] = data.get("premium_type")
        self.public_flags: Optional[int] = data.get("public_flags")

    def __str__(self):
        return f"{self.username}#{self.discriminator}"

    def __repr__(self):
        return f"<User id={self.id} username='{self.username}' bot={self.bot}>"    

    @property
    def mention(self) -> str:
        """The mention for this user"""
        return f"<@!{self.id}>"

    @property
    def avatar_url(self) -> Optional[str]:
        """The avatar URL for this user"""
        if self.avatar_hash:
            return f"{pycordia.cdn_url}/avatars/{self.id}/{utils.add_ext(self.avatar_hash)}"

    @property
    def banner_url(self) -> Optional[str]:
        """The banner URL for this user"""
        if self.banner_hash:
            return f"{pycordia.cdn_url}/banners/{self.id}/{utils.add_ext(self.banner_hash)}"

    @property
    def created_on(self) -> datetime:
        """The date and time the user was created on"""
        return utils.snowflake_to_date(int(self.id))

    @classmethod
    async def from_id(cls, user_id: str, use_cache: bool = True):
        """
        Get a User object, given a user ID

        Args:
            user_id (str): ID of the user to fetch
            use_cache (bool): Whether to check in cache first, or not

        Returns: A `pycordia.models.user.User` object
        """
        client = pycordia.models.active_client
        if not client:
            raise Exception("No initialized client found")

        # Check if its present in the cache first
        user = client.user_cache.get(user_id)
        if user and use_cache:
            return user

        # Otherwise, fetch directly from the API
        rs = await api.request("GET", f"users/{user_id}")
        user = User(rs)

        # Add to cache
        if len(client.user_cache.keys()) >= client.cache_size:
            first_user = list(client.user_cache.keys())[0]
            del client.user_cache[first_user]
        
        client.user_cache[user.id] = user

        return user

    def __eq__(self, other: 'User'):
        return self.id == other.id



import typing
from .models import Member, User, Message
from . import utils

class ReadyEvent:
    """
    Event called when the client is ready.


    Attributes:
        gateway_version: The version used for the WebSockets gateway
        user: The bot using the gateway
        guilds: A list of guild IDs the bot is in
        session_id: The session ID for the WebSockets session
        shard: The number of shards and their ID for this session
        partial_application: An Application object with an ID and flags
    """
    def __init__(self, data: dict):
        self.gateway_version: int = data["v"]
        self.user: User = User(data["user"])
        self.guilds: typing.List[int] = [int(guild["id"]) for guild in data ["guilds"]]
        self.session_id: str = data["session_id"]
        self.shard: tuple = data.get("shard", ())

        self.partial_application: dict = data["application"]


class MessageDeleteEvent:
    """Event called when message(s) deleted individually or in bulk
        
    Attributes:
        message_ids: The IDs of the message
        channel_id: The ID of the channel
        guild_id: The ID of the guild
        bulk: Whether the deletion was performed in bulk
        cached_message: The message, if it was cached by Pycordia
    """     
    def __init__(self, data: dict, bulk: bool, messages):
  
        if bulk:
            self.message_ids = data["ids"]
        else:
            self.message_ids = [data["id"]]

        self.cached_messages: typing.List[Message] = messages
        self.channel_id = data["channel_id"]
        self.guild_id = data.get("guild_id")
        self.bulk = bulk


class TypingStartEvent:
    """Event called when an user starts typing a message

    Attributes:
        timestamp: The ISO8601 timestamp for the message
        member: The member that started typing
        user_id: The ID of the member
        channel_id: The ID of the channel where event was registered
        guild_id: The ID of the guild where event was registered
    """

    def __init__(self, data: dict):
        self.timestamp = data["timestamp"]

        self.member = utils.make_optional(Member, data.get("member"))

        self.user_id = data["user_id"]
        self.channel_id = data["channel_id"]
        self.guild_id = data["guild_id"]




from __future__ import annotations
from datetime import datetime
import typing
import enum
from typing import List, Union, Optional

import pycordia
from .user import User
from pycordia import api, utils


class Emoji:
    """Represents server specific emojis."""
    def __init__(self, data: dict):
        self.emoji_id: Optional[str] = data.get("id")
        self.name: Optional[str] = data.get("name")
        self.roles: List[Role] = list(map(Role, data.get("roles", [])))
        self.user: Optional[User] = utils.make_optional(User, data.get("user", {}))
        self.requires_colons: Optional[bool] = data.get("require_colons")
        self.managed: Optional[bool] = data.get("managed")
        self.animated: Optional[bool] = data.get("animated")
        self.available: Optional[bool] = data.get("available")

    def __repr__(self):
        return f"<Emoji id={self.emoji_id} name='{self.name}'>"


class Member:
    """
    A Discord member part of a guild

    Attributes:
        user (User): The `pycordia.models.user.User` instance of the member
        nick (str): Nickname of the user. Can be `None` if no nickname has been set
        role_ids (List[Role]): The ID of the roles the user has
        joined_at (datetime): The time the user joined
        premium_since (datetime): The time the user started boosting the server
        deaf (bool): Whether the user has been deafened in a voice channel
        mute (bool): Whether the user has been muted in a voice channel
        pending (bool): Whether the user has not yet passed the guild's membership screening requirements
        permissions (string): Total permissions of the member in the channel, including overwrites

    Operations:
        - x == y: Checks if two channels are the same; returns None if a user is not present
        - repr(x): Returns a representation of this object
    """
    def __init__(self, data: dict):
        self.user: Optional[User] = utils.make_optional(User, data.get("user", {}))
        self.nick: Optional[str] = data.get("nick")
        self.guild_avatar_hash: Optional[str] = data.get("avatar")
        self.role_ids: list = data.get("roles", [])
        # NOTE: Uncomment when Roles are properly implemented
        # self.roles: list = list(map(Role, data.get("roles", [])))
        self.joined_at: datetime = datetime.fromisoformat(data["joined_at"])
        self.premium_since: Optional[datetime] = utils.make_optional(datetime.fromisoformat, data.get("premium_since"))

        self.deaf: bool = data["deaf"]
        self.mute: bool = data["mute"]

        self.pending: Optional[bool] = data.get("pending")
        self.permissions: Optional[str] = data.get("permissions")
    
    def __eq__(self, other) -> Optional[bool]:
        if self.user:
            return self.user.id == other.user.id

    def __repr__(self):
        return f"<Member user={str(self.user)}{' deaf'*self.deaf}{' mute'*self.mute}>"


class PartialGuild:
    """
    Represents a partial guild
    
    Attributes:
        id (str): ID of the guild.
        name (str): Name of the guild.
        icon (str): Hash of the guild's icon.
        owner (bool): Whether the connected user owns this guild.
        features (List[str]): List of features available to the guild.
    """
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.icon_hash: Optional[str] = data.get("icon")
        self.owner: Optional[bool] = data.get("owner")
        self.permissions: Optional[str] = data.get("permissions")
        self.features: List[str] = data.get("features", [])

    def __repr__(self):
        return f"<PartialGuild id={self.id} name='{self.name}'>"


class RoleTags:
    def __init__(self, data: dict):
        self.bot_id = data.get("bot_id")
        self.integration_id = data.get("integration_id")
        self.premium_subscriber = data.get("premium_subscriber")

    def __repr__(self):
        return f"<RoleTags bot_id={self.bot_id}>"


class Role:
    """
    Model to represent a guild's role.

    Attributes:
        id (str): ID of role
        name (str): Name of role
        color (int): Role color
        colour (int): Role colour
        hoist (bool): Whether the role is hoisted
        icon_hash (str): The icon hash for the icon
        unicode_emoji (str): The role's unicode emoji if a default emoji
        position (int): Role position in the hierarchy
        permissions (str): Permissions granted to the role
        managed (bool): Whether the role is a managed role
        mentionable (bool): Whether users can mention the role
        tags (List[RoleTags]): The tags for this role

    Operations:
        - x == y: Checks if two roles are the same
        - x > y: Checks if role `x` is higher up than role `y`
        - x < y: Checks if role `y` is higher up than role `x`
    """
    def __init__(self, data: dict):
        self.role_id: Union[str, None] = data.get("id")
        self.role_name: Union[str, None] = data.get("name")

        self.color: Union[int, None] = data.get("color")
        self.colour = self.color

        self.hoist: Union[bool, None] = data.get("hoist")
        self.position: Union[int, None] = data.get("position")
        self.permissions: Union[str, None] = data.get("permissions")
        self.managed: Union[bool, None] = data.get("managed")
        self.mentionable: Union[bool, None] = data.get("mentionable")

        self.tags: List[RoleTags] = [RoleTags(tag) for tag in data.get("tags", [])]
    

    def __eq__(self, other) -> bool:
        return self.role_id == other.role_id
    
    def __gt__(self, other) -> bool:
        """Check if the role is higher up than another"""
        return self.role_id > other.role_id

    def __lt__(self, other) -> bool:
        """Check if the role is lower than another"""
        return self.role_id < other.role_id

    def __repr__(self):
        return f"<Role id={self.role_id} name='{self.role_name}'>"


class GuildVerificationLevel(enum.Enum):
    none = 0
    low = 1
    medium = 2
    high = 3
    very_high = 4

class GuildNotificationLevel(enum.Enum):
    all_messages = 0
    only_mentions = 1

class GuildExplicitContentFilter(enum.Enum):
    disabled = 0
    members_without_roles = 1
    all_members = 2


class GuildNSFWLevel(enum.Enum):
    default = 0
    explicit = 1
    safe = 2
    age_restricted = 3

class GuildMFALevel(enum.Enum):
    none = 0
    elevated = 1

class GuildPremiumTier(enum.Enum):
    none = 0
    tier_1 = 1
    tier_2 = 2
    tier_3 = 3

class Guild:
    def __init__(self, data: dict) -> None:
        self.id: str = data["id"]
        self.name: str = data["name"]
        self.icon_hash: Optional[str] = data.get("icon")
        self.template_icon_hash: Optional[str] = data.get("icon_hash")
        self.splash_hash: Optional[str] = data.get("splash")
        self.discovery_splash_hash: Optional[str] = data.get("discovery_splash")
        self.is_bot_owner: Optional[str] = data.get("owner")
        self.owner_id: str = data["owner_id"]
        self.permissions: Optional[str] = data.get("permissions")
        self.afk_channel_id: Optional[str] = data.get("afk_channel_id")
        self.afk_timeout: int = data["afk_timeout"]
        self.has_widget_enabled: Optional[bool] = data.get("widget_enabled")
        self.widget_channel_id: Optional[str] = data.get("widget_channel_id")
        self.verification_level = GuildVerificationLevel(data["verification_level"])
        self.message_notifcations_level = GuildNotificationLevel(data["default_message_notifications"])
        self.explicit_content_filter = GuildExplicitContentFilter(data["explicit_content_filter"])
        self.roles: List[Role] = [Role(role) for role in data["roles"]]
        self.emojis: List[Emoji] = [Emoji(emoji) for emoji in data["emojis"]]
        self.features: List[str] = data["features"]
        self.mfa_level = GuildMFALevel(data["mfa_level"])
        self.application_id: Optional[str] = data.get("application_id")
        self.system_channel_id: Optional[str] = data.get("system_channel_id")
        self.system_channel_flags = data.get("system_channel_flags")
        self.rules_channel_id: Optional[str] = data.get("rules_channel_id")
        self.joined_at: Optional[str] = data.get("joined_at")
        self.large: Optional[bool] = data.get("large")
        self.unavailable: Optional[bool] = data.get("unavailable")
        self.member_count: int = data.get("member_count", -1)
        self.voice_states = data.get("voice_states")
        self.members: List[Member] = [Member(mem) for mem in data.get("members", [])]
        self.channels = data.get("channels")
        self.threads = data.get("threads")
        self.presences = data.get("presences")
        self.max_presences: Optional[int] = data.get("max_presences")
        self.max_members: Optional[int] = data.get("max_members")
        self.vanity_url_code: Optional[str] = data.get("vanity_url_code")
        self.description: Optional[str] = data.get("description")
        self.banner_hash: Optional[str] = data.get("banner")
        self.premium_tier = GuildPremiumTier(data["premium_tier"])
        self.premium_subscription_count: Optional[int] = data.get("premium_subscription_count")
        self.preferred_locale: str = data["preferred_locale"]
        self.public_updates_channel_id: Optional[str] = data.get("public_updates_channel_id")
        self.max_video_channel_users: Optional[int] = data.get("max_video_channel_users")
        self.approximate_member_count: Optional[int] = data.get("approximate_member_count")
        self.approximate_presence_count: Optional[int] = data.get("approximate_presence_count")
        self.welcome_screen = data.get("welcome_screen")
        self.nsfw_level = GuildNSFWLevel(data["nsfw_level"])
        self.stage_instances = data.get("stage_instances")
        self.stickers = data.get("stickers")

    @classmethod
    async def from_id(cls, guild_id: str, *, with_counts: bool = False) -> Guild:
        """Get a guild from an ID
        
        Args:
            guild_id (str): A guild ID
            with_counts (bool): Whether or not to add approximate member and presence counts
        
        Returns: A `pycordia.models.guild.Guild` object
        """
        rs = await api.request("GET", f"guilds/{guild_id}{'?with_counts=true'*with_counts}")
        return Guild(rs)



import pycordia
import aiohttp

async def request(method: str, endpoint: str, *, json_data={}):
    client = pycordia.models.active_client
    if not client:
        raise Exception("No initialized client found.")
    
    async with aiohttp.ClientSession(headers={
        "Authorization": f"Bot {client.ws.bot_token}",
        "Accept": "application/json"
    }) as session:
        kws = { "json": json_data } if json_data else {}

        async with session.request(method, 
            f"{pycordia.api_url}/{endpoint}", **kws
        ) as resp:
            rs = await resp.json()
            
            if not resp.ok:
                raise pycordia.errors.QueryError(rs.get("code"), rs.get("message"))
            
            return rs


import typing
import inspect
import datetime

def obj_from_dict(data: dict, obj: typing.Callable, alias: dict = None):
    """Convert a dictionary `data` into an `obj`.
    
    Parameters:
        data (dict): The data to provide.
        obj (typing.Callable): The object created from `data`
        alias (dict): A dictionary containing aliases for any parts of the dictionary. \
            Follows a (param-alias) pair
    """
    sig = inspect.signature(obj)
    new_inf = {}
    for param in sig.parameters:
        if alias and param in alias:
            new_inf[param] = data[alias[param]]
        elif param in data:    
            new_inf[param] = data[param]

    return obj(**new_inf)

def obj_to_dict(obj: typing.Any, alias: dict = None):
    """Convert an `obj` into a dictionary.
    
    Parameters:
        obj (typing.Any): The object to convert
        alias (dict): A dictionary containing aliases for any parts of the object. \
            Follows a (attribute-alias) pair
    """
    new_inf = {}
    if obj.__dict__:
        for key, value in obj.__dict__.items():
            if not key.startswith("_"):
                if alias and key in alias:
                    new_inf[alias[key]] = value
                else:
                    new_inf[key] = value

    return new_inf        

def mutually_exclusive(*argument_names):
    def factory(fun):
        async def wrapper(*args, **kwargs):
            mutuals = []
            for kw in kwargs:
                if kw in argument_names:
                    mutuals.append(kw)

            if len(mutuals) > 1:
                raise Exception(f"Only one of this group {argument_names} can be provided at a time, found {len(mutuals)}.")

            return await fun(*args, **kwargs)
        return wrapper
    return factory

def snowflake_to_date(snowflake: int) -> datetime.datetime:
    """Converts a snowflake to a valid `datetime.datetime` object

    Args:
        snowflake (int): A valid snowflake
    
    Returns: `datetime.datetime`
    """
    
    # The algorithm applied is the same as documented in
    # https://discord.com/developers/docs/reference

    # (snowflake >> 22) + Discord Epoch
    ms = (snowflake >> 22) + 1420070400000
    
    # Divided by 1000 and then provided to datetime.datetime
    return datetime.datetime.utcfromtimestamp(ms / 1000)

def make_optional(callable: typing.Callable, *args, **kwargs) -> typing.Optional[typing.Any]:
    """Return the result of `callable` if `args` or `kwargs` evaluate to True, else None

    Args:
        callable (typing.Callable): A callable object
    """
    if any(bool(arg) for arg in args) or \
        any(bool(kwargs[kwarg]) for kwarg in kwargs):
        return callable(*args, **kwargs)

def add_ext(hash_: str):
    if hash_.startswith("a_"):
        return f'{hash_}.gif'
    return f'{hash_}.png'



class QueryError(Exception):
    """General error raised by the HTTP API"""
    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__(f"{message} (error code {code})")


class GatewayError(Exception):
    """General error raised by the gateway"""
    def __init__(self, code, message) -> None:
        extra = None

        # 4001 = Unknown opcode
        # 4002 = Decode error
        # 4007 = Invalid sequence
        # 4012 = Invalid API version
        if code in (4001, 4002, 4007, 4012):
            extra = "This is an internal error."
        # Not authenticated
        elif code == 4003:
            extra = "Please authenticate and try again."
        # Authentication error
        elif code == 4004:
            extra = "Make sure the token you provided is valid and has not expired."
        # Rate limited
        elif code == 4008:
            extra = "Please wait a few seconds, then try again."
        # Invalid intents
        elif code == 4013:
            extra = "Make sure you have provided a valid intent value."
        # Disallowed intents
        elif code == 4014:
            extra = "Make sure you have approved/enabled the use of the provided intents."
        
        super().__init__(f"{message} {extra} (error code {code})")
    

# NOTE: For future use
class ComponentError(Exception):
    pass


class MutuallyExclusiveError(Exception):
    """Raised when multiple mutually exclusive values are provided to a function"""
    def __init__(self, args: tuple, mutual_count: int) -> None:
        super().__init__(
            f"Only one of this group {args} "
            f"can be provided at a time, found {mutual_count}."    
        )



import aiohttp
import enum
from .models.message import Member
from typing import Optional

class AppCommandOptionTypes(enum.Enum):
    sub_command = 1
    sub_command_group = 2
    string = 3
    integer = 4
    boolean = 5
    user = 6
    channel = 7
    role = 8
    mentionable = 9
    number = 10

class AppCommandInteractionDataOption:
    def __init__(self, data: dict):
        self.name: str = data["name"]
        self.command_type = AppCommandOptionTypes(int(data["type"]))
        self.value: Optional[str] = data.get("value")
        self.options = data.get("options", [])

    def add_option(self, *, name: str, value: str, command_type):
        self.options.append(AppCommandInteractionDataOption(
            {
                "name": name,
                "value": value,
                "type": command_type
            }
        ))


class AppCommandOption:
    def __init__(self, data: dict):
        self.option_type = AppCommandOptionTypes(data["type"])
        self.name: str = data["name"]
        self.description: str = data["description"]
        self.required: bool = data.get("required", False)
        self.choices: list = data.get("choices", [])
        self.options: list = data.get("options", [])

    def add_choice(self, *, name: str, value: str):
        self.choices.append({
            "name": name,
            "value": value
        })
    
    def add_option(self, *, option_type, name: str, description: str, required: bool = False):
        self.options.append(AppCommandOption(
            {
                "type": option_type,
                "name": name,
                "description": description,
                "required": required
            }
        ))


class AppCommand:
    class CommandTypes(enum.Enum):
        chat_input = 1
        user = 2
        message = 3

    def __init__(self, data: dict):
        CommandType = AppCommand.CommandTypes

        self.user_id: str = data["id"]
        if data.get("type"):
            self.command_type: Optional[CommandType] = CommandType(data["type"])
        else:
            self.command_type: Optional[CommandType] = None
        self.application_id: str = data["application_id"]
        self.guild_id: Optional[str] = data.get("guild_id")
        self.name: str = data["name"]
        self.description: str = data["description"]
        self.default_permission: bool = data.get("default_permissions", True)


class InteractionClient:
    registered_commands = {

    }
    @classmethod
    def slash_command(cls, *, name: str):
        def command(func):
            cls.registered_commands[name] = func
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return command


# interact = InteractionClient()

# @interact.slash_command(
#     name="blep", description="Send a random adorable animal photo",
#     options=[AppCommandOption]
# )
# async def blep(*args, **kwargs):
#     print(args, kwargs)





from __future__ import annotations
from pycordia import api

import enum
import datetime


class StatusType(enum.Enum):
    online = "online"
    dnd = "dnd"
    idle = "idle"
    invisible = "invisible"
    offline = "offline"


class Presence:
    def __init__(self, *, 
        since: datetime.datetime, activities,
        status: StatusType, afk: bool = True    
    ):
        pass